@@ -7,12 +7,20 @@ taskhawk
 
 
 # semantic versioning (http://semver.org/)
-VERSION = '3.1.0'
+VERSION = '4.0.0'
 
 
-from .models import Metadata, Priority  # noqa
-from .task_manager import AsyncInvocation, task, Task  # noqa
-from .consumer import listen_for_messages, process_messages_for_lambda_consumer  # noqa
+try:
+    from .backends.aws import AWSMetadata  # noqa
+except ImportError:
+    pass
+try:
+    from .backends.gcp import GoogleMetadata  # noqa
+except ImportError:
+    pass
 from .commands import requeue_dead_letter  # noqa
+from .consumer import listen_for_messages, process_messages_for_lambda_consumer  # noqa
 from .exceptions import *  # noqa
+from .models import Metadata, Priority  # noqa
 from .publisher import publish  # noqa
+from .task_manager import AsyncInvocation, task, Task  # noqa
